#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Feb 22, 2011 13:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_annotdict(file, afile, ic):
	#####################################################################
	# put annotations from afile into dictionary, and return dictionary
	#####################################################################
	
	ann={}
	i=-1
	while True :
		line = afile.readline()
		if len(line) == 0:
			# EOF
			break
		line = line.strip()
		cols = line.split('\t')
		if len(cols) < 2:
			# end of table
			break
		i+=1
		chrpos = cols[ ic[0] ] [3:] + '_' + cols[ ic[1] ]
		# if cols[ic[1]] == "152644604":
		#	print chrpos
		#	 1_152644604
		try:
			snp_name = cols[ ic[2] ]
			ref = cols[ ic[3] ]
			gene_name1 = cols[ ic[4] ]
			function = cols[ ic[5] ]
			effect = cols[ ic[6] ]
			grantham = cols [ ic[7] ] 
			g1000ceu = cols[ ic[8] ]
			omim = cols[ ic[9] ]
			hgmd_disease = cols[ ic[10] ]
			pph2_prediction = cols[ ic[11] ]
			pph2_prob = cols[ ic[12] ].replace(',','.')
			snap_prediction = cols[ ic[13] ]
			sift_prediction = cols[ ic[14] ]
			sift_score = cols[ ic[15] ]
		except:
			print ' *** Warning: possible formatting error in %s at chr %s pos %' % (file, ic[0], ic[1])
			snp_name = ''
			ref = ''
			gene_name1 = ''
			function = ''
			effect = ''
			grantham = ''
			g1000ceu = ''
			omim = ''
			hgmd_disease = ''
			pph2_prediction = ''
			pph2_prob = ''
			snap_prediction = ''
			sift_prediction = ''
			sift_score = ''
		ann[chrpos] = (snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim,
		 hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, sift_score)
	# print ann
	return ann

def pib_annot_read (infile, inf, ft):
	#####################################################################
	# read and split a single line from infile. 11/11/11
	#####################################################################
	while True:
		line = inf.readline()
		if len(line) == 0:
			# EOF
			cols=[]
			ncols=0
			chrpos=''
			break
		else:
			# test for empty or comment lines:
			cols=line.split('\t')
			ncols = len(cols)
			if ncols > 9:
				break
	if ncols > 0:
		if ft == 1:
			# pibase_consensus
			chrpos = cols[0] + '_' + cols[1]
		elif ft == 2:
			# pibase_fisherdiff
			#chrpos = cols[7] + '_' + cols[8]
			chrpos = cols[8] + '_' + cols[9]
		elif ft == 3:
			# pibase_diff
			chrpos = cols[1] + '_' + cols[2]
	return chrpos, cols, ncols

def pib_annot_write_c (inf, out, ann, annot, chrpos, cols, ncols):
	#####################################################################
	# merge pibase_consensus line with annotations and write to outfile
	#####################################################################
	# snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim, hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, sift_score = ann[chrpos]
	try:
		snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim, hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, sift_score = ann[chrpos]
	except:
		# have to change this ?
		# print 'ann', ann
		sys.exit (' *** Error: chr%s not defined in file %s' % (chrpos, annot))
	
	# build line for outfile:
	outline = '%s\t%s' % (gene_name1, snp_name)
	outline += '\t%s\t%s\t%s\t%s\t%s' % (cols[0],  cols[1],   function, effect, g1000ceu)
	outline += '\t%s\t%s\t%s\t%s\t%s' % (grantham, omim, hgmd_disease, pph2_prediction, pph2_prob)
	outline += '\t%s\t%s\t%s' % (snap_prediction, sift_prediction, sift_score)
	for i in range(2,ncols):
		outline += '\t'+cols[i]
	out.write(outline)

def pib_annot_write_fd (inf, out, ann, annot, chrpos, cols, ncols):
	#####################################################################
	# merge pibase_fisherdiff line with annotations and write to outfile
	#####################################################################
	try:
		snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim, \
		hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, \
		sift_score = ann[chrpos]
	except:
		# have to change this ?
		sys.exit (' *** Error: chr%s not defined in file %s' % (chrpos, annot))
	
	# build line for outfile:
	outline = cols[0]
	#for i in range(1,7):
	for i in range(1,8):
		outline += '\t'+cols[i]
	#outline += '\t%s\t%s\t%s\t%s' % (gene_name1, snp_name, cols[7], cols[8])
	outline += '\t%s\t%s\t%s\t%s' % (gene_name1, snp_name, cols[8], cols[9])
	outline += '\t%s\t%s\t%s' % (function, effect, g1000ceu)
	outline += '\t%s\t%s\t%s\t%s\t%s' % (grantham, omim, hgmd_disease, pph2_prediction, pph2_prob)
	outline += '\t%s\t%s\t%s' % (snap_prediction, sift_prediction, sift_score)
	#for i in range(9,ncols):
	for i in range(10,ncols):
		outline += '\t'+cols[i]
	out.write(outline)

def pib_annot_write_d (inf, out, ann, annot, chrpos, cols, ncols):
	#####################################################################
	# merge pibase_diff line with annotations and write to outfile
	#####################################################################
	try:
		snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim, \
		hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, \
		sift_score = ann[chrpos]
	except:
		# have to change this ?
		sys.exit (' *** Error: chr%s not defined in file %s' % (chrpos, annot))
	
	# build line for outfile:
	outline = cols[0]
	outline += '\t%s\t%s\t%s\t%s' % (gene_name1, snp_name, cols[1], cols[2])
	outline += '\t%s\t%s\t%s' % (function, effect, g1000ceu)
	outline += '\t%s\t%s\t%s\t%s\t%s' % (grantham, omim, hgmd_disease, pph2_prediction, pph2_prob)
	outline += '\t%s\t%s\t%s' % (snap_prediction, sift_prediction, sift_score)
	for i in range(3,ncols):
		outline += '\t'+cols[i]
	out.write(outline)

def pib_annot (ft, annot1, annot2, infile, outfile):
	'''        #######################################################################################
	#  pibase_annot.py  v1.4.3
	#######################################################################################
	#  Adds snp annotations to a pibase file (see type). 
	#
	#  Usage:	pibase_annot.py <type> <annot1> [annot2] <infile> <outfile>
	#  where
	#  type:	type of input file 
	#               (1=pibase_consensus; 2=pibase_fisherdiff; 3=pibase_diff)
	#  annot1:	snpacts-annotations file 1 (control sample, created by outputact)
	#  annot2:	snpacts-annotations file 2 (case sample, created by outputact)
	#		annot2 is not specified for pibase_consensus (type=1)
	#  infile:	tab-separated file 
	#		(from pibase_diff, pibase_fisherdiff or pibase_consensus)
	#  outfile:	results file with merging of annotations and infile
	#######################################################################################
	#  Author: Michael Forster,                            m.forster@ikmb.uni-kiel.de         
	#  Date:   22 February 2011, 29-30 March 2011, 20 April 2011, 18 May 2011, 11/11/11
	#######################################################################################
	'''

	############################################################################
	#  read snp annotations into dictionary
	############################################################################

	afile = open (annot1, 'r')
	
	# find table header
	while True :
		line = afile.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (annot1))
		if line.startswith('#chrom\tposition\tsnp_name'):
			header = line
			break
	
	# find columns of interest:
	header = header.strip()
	# print header
	cols=header.split('\t')
	c=[]
	c.append('#chrom')
	c.append('position')
	c.append('snp_name')
	c.append('refBase')
	c.append('gene symbol')
	c.append('cathegory')
	c.append('effect')
	c.append('Grantham')
	c.append('g1000CEU')
	c.append('OMIM')
	c.append('HGMD_disease')
	c.append('pph2_prediction')
	c.append('pph2_prob')
	c.append('snap_Prediction')
	c.append('sift_Prediction')
	c.append('sift_score')
	# 0 c.append('#chrom')
	# 1 c.append('position')
	# 2 c.append('snp_name')
	# 3 c.append('refBase')
	# 4 c.append('gene symbol')
	# 5 c.append('cathegory')
	# 6 c.append('effect')
	# 7 c.append('Grantham')
	# 8 c.append('g1000CEU')
	# 9 c.append('OMIM')
	# 10 c.append('HGMD_disease')
	# 11 c.append('pph2_prediction')
	# 12 c.append('pph2_prob')   Here we need to string.replace (',','.')
	# 13 c.append('snap_Prediction')
	# 14 c.append('sift_Prediction')
	# 15 c.append('sift_score')
	ic=[]
	maxcol=0
	for i in c:
		# print i
		n=0
		for col in cols:
			if i==col:
				# print 'found '+col+' in column '+str(n)
				ic.append(n)
				maxcol=max(maxcol,n)
				break
			n+=1
	if len(ic) != len(c):
			print ' *** Error: 1 or more annotation columns of interest not found' 
			print  c
			print ic
			sys.exit (' *** Aborted' )
	# print 'ic',ic


	#d=open ('debug.txt','w')
	## Put annotations into list:
	
	#####################################################################
	# put annotations into dictionary:
	#####################################################################

	ann1={}
	ann1 = pib_annotdict(annot1, afile, ic) 
	afile.close()

	#print 'STEP 2' 
	if ft == 2 or filetype == 3:
		
		afile = open (annot2, 'r')
		# find table header
		while True :
			line = afile.readline()
			#print line
			if len(line) == 0:
				# EOF
				sys.exit (' *** Error: End of file %s reached before table header could be found' % (annot2))
			if line.startswith('#chrom\tposition\tsnp_name'):
				header = line
				break
		## later : split header line and check columns ic against c
		ann2={}
		ann2 = pib_annotdict(annot2, afile, ic) 
		afile.close()
		
	############################################################################
	# open infile and outfile and copy infile header lines to outfile:
	############################################################################
	
	inf = open (infile, 'r')
	out = open (outfile, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_annot v1.4.3 on '+ outtime)
	if ft == 2 or ft == 3:
		out.write ('# using annotation file 1: %s\n' % (annot1))
		out.write ('#   and annotation file 2: %s\n' % (annot2))
	else:
		out.write ('# using annotation file: %s\n' % (annot1))
	out.write ('# from: %s\n' % (infile))

	n = 0
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile))
		if line.startswith('#') and (line.find('\tConsensus Genotype\t')>1 or line.find('\tBest\t')>1):
			# Pad the two first table header lines (which are above the 'real' header line)
			cols=line.split('\t')
			outline=cols[0]
			for i in range(0,13):
				outline += '\t'
			for i in range(1,len(cols)):
				outline += '\t' + cols[i]
			out.write (outline)
		elif line.startswith('# C\tPos[1]\t') and ft == 1:
			# pibase_consensus
			header = line
			break
		elif line.startswith('#D\tD_error\tp.med\t') and ft == 2:
			# pibase_fisherdiff
			header = line
			break
		elif line.startswith('#D\tC\tPos[1]\tRef') and ft == 3:
			# pibase_diff
			header = line
			break
		else:
			out.write(line)
		
	############################################################################
	# table header line for outfile
	############################################################################
	cols = header.split('\t')
	ncols = len(cols)
	if ft == 1:
		# pibase_consensus
		outheader = '#Gene\trsID\tChr\tPos[1]\tFunction\tEffect\tg1000CEU'+ \
		'\tGrantham\tOMIM\tHGMD\tPph2_pred\tPph2_prob\tSnap\tSift_pred\tSift_score'
		for i in range(2,ncols):
			outheader += '\t'+cols[i]
	elif ft == 2:
		# pibase_fisherdiff
		outheader = '#D'
		#for i in range(1,7):
		for i in range(1,8):
			outheader += '\t'+cols[i]
		outheader += '\tGene\trsID\tChr\tPos[1]\tFunction\tEffect\tg1000CEU'+ \
		'\tGrantham\tOMIM\tHGMD\tPph2_pred\tPph2_prob\tSnap\tSift_pred\tSift_score'
		# for i in range(9,ncols):
		for i in range(10,ncols):
			outheader += '\t'+cols[i]
	elif ft == 3:
		# pibase_diff
		# snp_name, ref, gene_name1, function, effect, grantham, g1000ceu, omim, 
		# hgmd_disease, pph2_prediction, pph2_prob, snap_prediction, sift_prediction, sift_score
		outheader = '#D\tGene\trsID\tChr\tPos[1]\tFunction\tEffect\tg1000CEU'+ \
		'\tGrantham\tOMIM\tHGMD\tPph2_pred\tPph2_prob\tSnap\tSift_pred\tSift_score'
		for i in range(3,ncols):
			outheader += '\t'+cols[i]
			
	out.write(outheader)

	############################################################################
	# loop over remaining lines to annotate these
	############################################################################

	if ft == 1:
		# pibase_consensus
		while True:
			n += 1
			# read and split a single line from infile
			chrpos, cols, ncols = pib_annot_read (infile, inf, ft)
			if ncols == 0:
				break
			# merge line with annotations and write to outfile
			pib_annot_write_c (inf, out, ann1, annot1, chrpos, cols, ncols)
	elif ft == 2:
		# pibase_fisherdiff
		while True:
			n += 1
			# read and split a single line (control sample) from infile
			chrpos, cols, ncols = pib_annot_read (infile, inf, ft)
			if ncols == 0:
				break
			# merge line with annotations and write to outfile
			pib_annot_write_fd (inf, out, ann1, annot1, chrpos, cols, ncols)
			
			n += 1
			# read and split a single line (case sample) from infile
			chrpos, cols, ncols = pib_annot_read (infile, inf, ft)
			if ncols == 0:
				break
			# merge line with annotations and write to outfile
			pib_annot_write_fd (inf, out, ann2, annot2, chrpos, cols, ncols)
	elif ft == 3:
		# pibase_diff
		while True:
			n += 1
			# read and split a single line from infile
			chrpos, cols, ncols = pib_annot_read (infile, inf, ft)
			if ncols == 0:
				break
			# merge line with annotations and write to outfile
			pib_annot_write_d (inf, out, ann1, annot1, chrpos, cols, ncols)
			
			if cols[0] != '=' and ncols != 0: 
				# read and split a single line (case sample with genotype difference)
				chrpos, cols, ncols = pib_annot_read (infile, inf, ft)
				if ncols == 0:
					break
				# merge line with annotations and write to outfile
				pib_annot_write_d (inf, out, ann2, annot2, chrpos, cols, ncols)
	inf.close()
	out.close()

##############################################
# the actual main
##############################################

nargs = len(sys.argv)
if nargs < 5:
	print pib_annot.__doc__
	sys.exit()
else:
	filetype = int(sys.argv[1])
	annot1 = sys.argv[2]
	annot2 = ''
	k=2
	if filetype != 1:
		if nargs == 5:
			print pib_annot.__doc__
			sys.exit()
		annot2 = sys.argv[3]
		k=3
	infile = sys.argv[k+1]
	outfile = sys.argv[k+2]

setlimits()

pib_annot (filetype, annot1, annot2, infile, outfile)
print('Done')
