#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Sep 21, 2011 13:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_gfs_dict(file, afile, ic, swapper):
	#####################################################################
	# put annotations from afile into dictionary, and return dictionary
	#####################################################################
	
	ann={}
	i=-1
	while True :
		line = afile.readline()
		if len(line) == 0:
			# EOF
			break
		line = line.strip()
		cols = line.split('\t')
		if len(cols) < 2:
			# end of table
			break
		i+=1
		chrnum = cols[ ic[0] ] [3:]
		if chrnum == 'X':
			chrnum = '23'
		elif chrnum == 'Y':
			chrnum = '24'
		chrpos = chrnum + '_' + cols[ ic[1] ]
		# if cols[ic[1]] == "152644604":
		#	print chrpos
		#	 1_152644604
		try:
			#refbase = cols[ ic[2] ]
			# convert to upper case, because ref base may be lower case in repeat masked ucsc fasta references
			refbase = cols[ ic[2] ].upper()
			varbase = cols[ ic[3] ]
			muttype = cols[ ic[4] ]
		except:
			print ' *** Warning: possible formatting error in %s at chr %s pos %' % (file, ic[0], ic[1])
			refbase = ''
			varbase = ''
			muttype = ''
		if muttype == 'hom':
			gen = varbase + varbase
		else:
			gen = refbase + varbase
			# swap e.g. CA to AC, etc:
			gen = swapper[gen]
		ann[chrpos] = gen
	#print ann
	return ann

def pib_gfs (ft, annot1, annot2, infile, outfile, hdr):
	'''        #######################################################################################
	#  pibase_gen_from_snpacts.py  v1.4.3
	#######################################################################################
	#  Inserts snpacts genotypes into a pibase file (see type), before "BestGen" column.
	#  Note that multiple runs of this script can be performed to insert the genotypes
	#  from several different snp-callers (samtools/Bioscope/GATK).
	#  Only chromosomes chr1-chr22, chrX, chrY are recognised in snpacts files.
	#
	#  Usage:  pibase_gen_from_snpacts <type> <s1> [s2] <in> <out> <hdr>
	#  where
	#  type:  type of input file 
	#         (1=pibase_consensus; 
	#         not implemented:   2=pibase_fisherdiff; 3=pibase_diff)
	#  s1:    snpacts file 1 (diff: control sample), created by outputact
	#  s2:    snpacts file 2 (diff: case sample), created by outputact
	#         s2 is not specified for pibase_consensus (type=1)
	#  in:    tab-separated file 
	#         (from pibase_diff, pibase_fisherdiff or pibase_consensus)
	#  out:   results file with merging of annotations and infile
	#  hdr:   header of inserted snpacts-column (e.g. samtools, Bioscope, BWA)
	#######################################################################################
	#  Author: Michael Forster,                            m.forster@ikmb.uni-kiel.de
	#  Date:   21 September 2011, 10 November 2011
	#######################################################################################
	'''
	if ft==2 or ft == 3:
		sys.exit ('*** Sorry: Only pibase_consensus is currently implemented. Run aborted')
	
	############################################################################
	#  read snp annotations into dictionary
	############################################################################

	afile = open (annot1, 'r')
	
	# find table header
	while True :
		line = afile.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (annot1))
		if line.startswith('#chrom\tposition\tsnp_name'):
			header = line
			break
	
	# find columns of interest:
	header = header.strip()
	# print header
	cols=header.split('\t')
	c=[]
	c.append('#chrom')
	c.append('position')
	c.append('refBase')
	c.append('varBase')
	c.append('mutType')
	# 0 c.append('#chrom')
	# 1 c.append('position')
	# 2 c.append('refBase')
	# 3 c.append('varBase')
	# 4 c.append('mutType')
	ic=[]
	maxcol=0
	for i in c:
		# print i
		n=0
		for col in cols:
			if i==col:
				# print 'found '+col+' in column '+str(n)
				ic.append(n)
				maxcol=max(maxcol,n)
				break
			n+=1
	if len(ic) != len(c):
			print ' *** Error: 1 or more annotation columns of interest not found' 
			print  c
			print ic
			sys.exit (' *** Aborted' )
	# print 'ic',ic


	#d=open ('debug.txt','w')
	## Put annotations into list:
	
	#####################################################################
	# put annotations into dictionary:
	#####################################################################
	
	swapper={}
	swapper ['AA'] = 'AA'
	swapper ['CC'] = 'CC'
	swapper ['GG'] = 'GG'
	swapper ['TT'] = 'TT'
	swapper [''] = ''
	
	swapper ['AC'] = 'AC'
	swapper ['CA'] = 'AC'
	swapper ['AG'] = 'AG'
	swapper ['GA'] = 'AG'
	swapper ['AT'] = 'AT'
	swapper ['TA'] = 'AT'
	
	swapper ['CG'] = 'CG'
	swapper ['GC'] = 'CG'
	swapper ['CT'] = 'CT'
	swapper ['TC'] = 'CT'
	
	swapper ['GT'] = 'GT'
	swapper ['TG'] = 'GT'

	ann1={}
	ann1 = pib_gfs_dict (annot1, afile, ic, swapper) 
	afile.close()

	#print 'STEP 2' 
	if ft == 2 or filetype == 3:
		
		afile = open (annot2, 'r')
		# find table header
		while True :
			line = afile.readline()
			#print line
			if len(line) == 0:
				# EOF
				sys.exit (' *** Error: End of file %s reached before table header could be found' % (annot2))
			if line.startswith('#chrom\tposition\tsnp_name'):
				header = line
				break
		## later : split header line and check columns ic against c
		ann2={}
		ann2 = pib_gfs_dict(annot2, afile, ic, swapper) 
		afile.close()
		
	############################################################################
	# open infile and outfile and copy infile header lines to outfile:
	############################################################################
	
	inf = open (infile, 'r')
	out = open (outfile, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out.write ('# File created by pibase_gen_from_snpacts v1.4.3 on '+ outtime)
	if ft == 2 or ft == 3:
		out.write ('# using annotation file 1: %s\n' % (annot1))
		out.write ('#   and annotation file 2: %s\n' % (annot2))
	else:
		out.write ('# using annotation file: %s\n' % (annot1))
	out.write ('# from: %s\n' % (infile))

	n = 0
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile))
		if line.startswith('# C\tPos[1]\t') and ft == 1:
			# pibase_consensus
			header = line
			break
		elif line.startswith('#D\tD_error\tp.med\t') and ft == 2:
			# pibase_fisherdiff
			header = line
			break
		elif line.startswith('#D\tC\tPos[1]\tRef') and ft == 3:
			# pibase_diff
			header = line
			break
		else:
			# add two tabs in the respective two header lines
			if line.find('MapQV>RQVmin and MM<=MMmax') != -1:
				line=line.replace('# ','# \t\t')
			if line.find('Unique start points (on +- strands)') != -1:
				line=line.replace('# ','# \t\t')
			# output header line
			out.write(line)
		
	############################################################################
	# table header line for outfile
	############################################################################
	cols = header.split('\t')
	ncols = len(cols)
	
	# find column with "BestGen":
	bgc = -1
	for c in cols:
		bgc += 1
		if c=="BestGen":
			break
	if bgc == -1:
		print '*** Error: No BestGen column in table header line of file ',infile 
		sys.exit()
	
	outheader=''
	for i in range(0,bgc):
		outheader += cols[i] + '\t'
	outheader += hdr + '_disc\t' + hdr
	for i in range(bgc,ncols):
		outheader += '\t'+cols[i]
	out.write(outheader)

	############################################################################
	# loop over remaining lines to annotate these
	############################################################################

	while True:
		n += 1
		# read and split a single line from infile
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols=line.split('\t')
		ncols = len(cols)
		
		# key for dictionary of snpacts genotypes
		chrpos = cols[ ic[0] ] + '_' + cols[ ic[1] ]
		try:
			sg = ann1 [chrpos]
		except:
			sg = ''
		
		# incorporate this genotype into the line:
		outline = cols[0]
		for i in range(1,bgc):
			outline += '\t'+cols[i]
		# output whether snpacts genotype is discordant to pibase-genotype, and
		# the snpacts-genotype:
		disc = '0'
		if sg != cols[bgc]:
			disc = '1'
		outline += '\t'+ disc + '\t' + sg
		for i in range(bgc,ncols):
			outline += '\t'+cols[i]
		# output the line (last column already has a LF):
		out.write(outline)

	inf.close()
	out.close()

##############################################
# the actual main
##############################################

nargs = len(sys.argv)
if nargs < 5:
	print pib_gfs.__doc__
	sys.exit()
else:
	filetype = int(sys.argv[1])
	annot1 = sys.argv[2]
	annot2 = ''
	k=2
	if filetype != 1:
		if nargs == 5:
			print pib_gfs.__doc__
			sys.exit()
		annot2 = sys.argv[3]
		k=3
	infile = sys.argv[k+1]
	outfile = sys.argv[k+2]
	hdr = sys.argv[k+3]

setlimits()

pib_gfs (filetype, annot1, annot2, infile, outfile, hdr)
print('Done')
