#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Oct 27, 2011 13:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_gfs_dict(file, afile, ic, swapper):
	#####################################################################
	# put annotations from afile into dictionary, and return dictionary
	#####################################################################
	
	ann={}
	i=-1
	while True :
		line = afile.readline()
		if len(line) == 0:
			# EOF
			break
		line = line.strip()
		cols = line.split('\t')
		if len(cols) < 2:
			# end of table
			break
		i+=1
		chrnum = cols[ ic[0] ] [3:]
		if chrnum == 'X':
			chrnum = '23'
		elif chrnum == 'Y':
			chrnum = '24'
		chrpos = chrnum + '_' + cols[ ic[1] ]
		# if cols[ic[1]] == "152644604":
		#	print chrpos
		#	 1_152644604
		try:
			refbase = cols[ ic[2] ]
			varbase = cols[ ic[3] ]
			muttype = cols[ ic[4] ]
		except:
			print ' *** Warning: possible formatting error in %s at chr %s pos %' % (file, ic[0], ic[1])
			refbase = ''
			varbase = ''
			muttype = ''
		if muttype == 'hom':
			gen = varbase + varbase
		else:
			gen = refbase + varbase
			# swap e.g. CA to AC, etc:
			gen = swapper[gen]
		ann[chrpos] = gen
	#print ann
	return ann

def pib_rdf_ref (infile, outfile, cov):
	'''        #######################################################################################
	#  pibase_rdf_ref.py  v1.4.5
	#  Generates a reference base pibase_consensus file which may be used as the
	#  reference for pibase_fisherdiff comparisons in a group of samples, e.g. 
	#  for a pibase_to_rdf run and subsequent phylogenetic network analysis.
	#######################################################################################
	#  pibase_rdf_ref in out [cov]
	# 
	#  in:    pibase_consensus file (any sample from the group of samples)
	#  out:   pibase_consensus file with genotypes taken from Ref column of in-file 
	#  cov:   coverage (optional. default=50. Divisible by two.)
	#######################################################################################
	#  Michael Forster, m.forster@ikmb.uni-kiel.de                           22 May 2012
	#######################################################################################
	'''
	
	print 'infile:         ',infile
	print 'outfile:        ',outfile
	print 'cov per strand: ',cov

	############################################################################
	# table header line, Ref base column
	############################################################################
	
	header = []
	
	# version 1.4.3 and older:
	h = '# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn' + \
	'\tBestGen\tBestQual\tA+-\tB+-' + \
	'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
	'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n'
	header.append(h)
	
	# new versions:
	h = '# C\tPos[1]\tRef\tRefSeqContext\tGC[%]\tClass\tIgn' + \
	'\tBestGen\tBestQual\tA+-\tB+-' + \
	'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
	'\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV\tSP\tCV'+ \
	'\tHomologous\tHypervariable'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-'+ \
	'\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\tA+\tA-\tC+\tC-\tG+\tG-\tT+\tT-\tN+\tN-\n'
	header.append(h)
	
	############################################################################
	# Dictionaries of coverage counts for output
	############################################################################
	
	cov1 = 2*cov
	covs1a = [cov1,0,0,0]
	covs1c = [0,cov1,0,0]
	covs1g = [0,0,cov1,0]
	covs1t = [0,0,0,cov1]
	covs1n = [0,0,0,0]
	
	covs2a = [cov,cov, 0,0, 0,0, 0,0, 0,0]
	covs2c = [0,0, cov,cov, 0,0, 0,0, 0,0]
	covs2g = [0,0, 0,0, cov,cov, 0,0, 0,0]
	covs2t = [0,0, 0,0, 0,0, cov,cov, 0,0]
	covs2n = [0,0, 0,0, 0,0, 0,0, 0,0]
	
	covs1 = {}
	
	covs1 ['A'] = covs1a
	covs1 ['C'] = covs1c
	covs1 ['G'] = covs1g
	covs1 ['T'] = covs1t
	covs1 ['N'] = covs1n
	
	covs1 ['a'] = covs1a
	covs1 ['c'] = covs1c
	covs1 ['g'] = covs1g
	covs1 ['t'] = covs1t
	covs1 ['n'] = covs1n

	covs2 = {}
	
	covs2 ['A'] = covs2a
	covs2 ['C'] = covs2c
	covs2 ['G'] = covs2g
	covs2 ['T'] = covs2t
	covs2 ['N'] = covs2n
	
	covs2 ['a'] = covs2a
	covs2 ['c'] = covs2c
	covs2 ['g'] = covs2g
	covs2 ['t'] = covs2t
	covs2 ['n'] = covs2n
	
	# Capital letters for reference base:
	
	capref = {}
	
	capref ['A'] = 'A'
	capref ['C'] = 'C'
	capref ['G'] = 'G'
	capref ['T'] = 'T'
	capref ['N'] = 'N'

	capref ['a'] = 'A'
	capref ['c'] = 'C'
	capref ['g'] = 'G'
	capref ['t'] = 'T'
	capref ['n'] = 'N'

	############################################################################
	# open infile and find header line
	############################################################################
	
	inf = open (infile, 'r')
	
	n = 0
	while True:
		n += 1
		line = inf.readline()
		if len(line) == 0:
			# EOF
			sys.exit (' *** Error: End of file %s reached before table header could be found' % (infile))
		if line in header:
			iheader = header.index(line)
			break

	#######################################################################################################
	#  Write header of out-file
	#######################################################################################################

	out = open(outfile,"w")
	out.write ('# Reference base file in pibase_consensus format\n')
	out.write ('# Generated by pibase_rdf_ref.py v1.4.5 from pibase_consensus file:\n')
	out.write ('# %s\n' % (infile))
	out.write('#\n')

	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n#"
	out.write ('# Date: '+ outtime+'\n')

	if iheader==0:
		# version 1.4.3 and older:
		c = '# \t\t\t\t\t\t' + \
		'\tConsensus Genotype\t\t\t\t\t\t\t\t\t' + \
		'\t\t' + \
		'\t\t' + \
		'\tReads\t\t\t\t\t\t\t\t\t' + \
		'\tMapQV>RQVmin and MM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tMM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tRead>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tBaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tNo QV-Filtering\n'
		out.write (c)
		
		c = '# \t\t\t\t\t\t' + \
		'\tBest\t' +\
		'\t\t' + \
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n'
		out.write (c)
	else:
		# new versions:
		c = '# \t\t\t\t\t\t' + \
		'\tConsensus Genotype\t\t\t\t\t\t\t\t\t' + \
		'\t\t' + \
		'\t\t' + \
		'\tReads\t\t\t\t\t\t\t\t\t' + \
		'\t\t' + \
		'\tMapQV>RQVmin and MM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tMM<=MMmax and Read>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tRead>MLmin and BaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tBaseQVs>=QVmin\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t' + \
		'\tNo QV-Filtering\n'
		out.write (c)
		
		c = '# \t\t\t\t\t\t' + \
		'\tBest\t' +\
		'\t\t' + \
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\tFilt4\t\tFilt3\t\tFilt2\t\tFilt1\t\tFilt0\t' +\
		'\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t' + \
		'\tUnique start points (on +- strands)\t\t\t\t\t\t\t\t\t\tCoverage\t\t\t\t\t\t\t\t\t\n'
		out.write (c)
	
	out.write (header[iheader])

	############################################################################
	# loop over remaining lines to get coord, ref, refseq-context
	############################################################################

	while True:
		n += 1
		# read and split a single line from infile
		line = inf.readline()
		if len(line) == 0:
			# EOF
			break
		cols=line.split('\t')
		ncols = len(cols)
		
		# Ref and BestGen:
		ref = capref [ cols[2] ]
		bg = ref + ref
		
		#### build output line:
		
		# first 6 columns from in-file:
		outline = cols[0]
		for i in range(1,6):
			outline += '\t'+cols[i]
		# Columns Ign, BestGen, BestQual, A+-, B+-
		outline += '\t0\t'+bg+'\t\t+-\t+-'
		
		# Genotype columns for all filter levels:
		for i in range(0,10):
			outline += '\t'+bg
		
		# Both-stranded allele counts:
		s = '%s:%s:%s:%s' % (covs1[ref][0], covs1[ref][1],covs1[ref][2],covs1[ref][3])
		# f4
		outline += '\t'+s
		outline += '\t'+s
		# f3
		outline += '\t'+s
		outline += '\t'+s
		# f2
		outline += '\t'+s
		outline += '\t'+s
		# f1
		outline += '\t'+s
		outline += '\t'+s
		# f0
		outline += '\t'+s
		outline += '\t'+s
		
		if iheader !=0 :
			# Columns for homologous and hypervariable loci:
			outline += '\t\t'
		
		# Per-strand allele counts:
		s = '' 
		for i in range(0,10):
			s += '\t' + str( covs2[ref][i] )
		# f4
		outline += s
		outline += s
		# f3
		outline += s
		outline += s
		# f2
		outline += s
		outline += s
		# f1
		outline += s
		outline += s
		# f0
		outline += s
		outline += s + '\n'

		#### output the line:
		out.write(outline)

	inf.close()
	out.close()

##############################################
# the actual main
##############################################

nargs = len(sys.argv)
if nargs < 3:
	print pib_rdf_ref.__doc__
	sys.exit()
else:
	infile = sys.argv[1]
	outfile = sys.argv[2]
	# per-strand coverage
	cov = 25
	if nargs > 3:
		try:
			cov = int(sys.argv[3])/2
		except:
			print pib_rdf_ref.__doc__
			sys.exit()
setlimits()

pib_rdf_ref (infile, outfile, cov)
print('Done')
