#!/bin/env python
##!/usr/bin/python

import pysam
import sys
import os.path
import time
import resource

###########################################################################################################
#
#  pibase - GET pOSITION iNFORMATION AT base POSITION OF INTEREST
#
###########################################################################################################

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	# usage = resource.getrusage(resource.RUSAGE_SELF)

	#print ' usage ='
	#print usage
	
	#for name, desc in [
	#('RLIMIT_CORE', 'core file size'),
	#('RLIMIT_CPU',  'CPU time'),
	#('RLIMIT_FSIZE', 'file size'),
	#('RLIMIT_DATA', 'heap size'),
	#('RLIMIT_STACK', 'stack size'),
	#('RLIMIT_RSS', 'resident set size'),
	#('RLIMIT_NPROC', 'number of processes'),
	#('RLIMIT_NOFILE', 'number of open files'),
	#('RLIMIT_MEMLOCK', 'lockable memory address'),
	#]:
	#	limit_num = getattr(resource, name)
	#	soft, hard = resource.getrlimit(limit_num)
	#	print 'Maximum %-25s (%-15s) : %20s %20s' % (desc, name, soft, hard)

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

	#sys.exit()


def pib_ref (refseqfile, tablefile, outfile, nflank):
	'''        #######################################################################################
	#  pibase_ref v0.9 
	#  Get reference sequence around coordinate (for Sanger-sequencing primer design)
	#######################################################################################
	#  pibase_ref  infile  ref outfile  [nflank]
	#
	#   infile    file with list of genomic coordinates:
	#             <chrom> TAB <1-based_pos_in_chr> EOL
	#
	#   ref       reference sequence file
	#
	#   outfile   fasta-formatted output file
	#
	#######################################################################################
	#  Michael Forster                                        m.forster@ikmb.uni-kiel.de         
	#  21/05/2012
	#######################################################################################
	'''

	#######################################################################################
	# open output file
	#######################################################################################

	out = open(outfile,"w")

	#######################################################################################
	# Get chromosome (or contig) name, and 0-based position of interest (poi):
	# Read multiple positions of interest from input file
	###################################################################################

	tabf=open (tablefile, 'r')
	while True:
		line=tabf.readline()
		if len(line) == 0:
			# EOF reached
			break
		if line.startswith('#'):
			continue
		line=line.strip()
		
		# Get chromosome_name and poi (0-based) from tab-separated file:
		try:
			cols = line.split('\t')
			chrom = cols[0]
			poi = int(cols[1].strip()) - 1
			break
		except:
			print '*** Error: Bad format in first SNP-line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line)
			sys.exit('***        Line should be: <chromosome_name> TAB <1_based_position_in_chrom>')
		
	#######################################################################################
	# open refseq file 
	#######################################################################################

	# read chromnames, lengths and binary starts (0-based) from fasta-index-file:
	faifile=refseqfile+'.fai'
	faifh=open(faifile)
	fai={}
	for line in faifh:
		cols=line.split('\t')
		fai[cols[0]]=(int(cols[1]),int(cols[2]))
		# fai.append((cols[0],cols[1],cols[2]))
	# print fai
	# l,j=fai['CHR1']
	# print l,j
	#sys.exit()
	
	#######################################################################################################
	#  Write HEADER of OUT FILE
	#######################################################################################################

	# out.write ('# Position Info (1-based coordinates) generated by pibase from file '+ sam_filename+'\n')
	out.write ('# Reference sequences for Sanger sequencing primer design generated by pibase_ref v0.9\n')
	out.write ('# Reference: '+ refseqfile+'\n')
	out.write ('# Genomic coordinates of interest taken from file: '+ tablefile+'\n')
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S")
	out.write ('# Number of flanking bases (upstream and downstream of coordinate): %s base-pairs\n' % (nflank))
	out.write ('# Date: '+ outtime+'\n')
	
	#######################################################################################################
	# Start of WHILE loop over all positions of interest
	#######################################################################################################

	while True:

		#######################################################################################
		#  read  fasta sequence [poi-nflank; poi + nflank] from refseq file
		#######################################################################################
	
		refseq=open(refseqfile, 'rb')
		# testing:
		# refseq.seek(6)
		# reads 7th character in file:
		# t=refseq.read(1)
		# reads 7th-8th characters in file:
		# t=refseq.read(2)
		# print t
		# sys.exit()
	
		# chrom length and start in refseq fasta file
		lmax,k = fai[chrom]
		# line length (excl. EOL) in fasta file (must be constant):
		refseq.seek(k)
		t=refseq.read(1000)
		ll=t.find('\n')
		# fragment start and end in chromosome:
		##j1=max(0,poi+1-nflank)
		j1=max(0,poi-nflank)
		j2=min(poi+nflank,lmax)
		#print ' j1, j2: %s %s' % (j1, j2)
		# fragment start and end in refseq fasta file:
		nn=int(j1/ll)
		rr= j1 % ll
		jj1=k+nn*(ll+1)+rr
		nn=int(j2/ll)
		rr= j2 % ll
		jj2=k+nn*(ll+1)+rr
		# number of bytes
		n=jj2-jj1+1
		# print jj1, jj2, n
		# read n bytes from position jj1, discarding the EOLs:
		# (as strip does not work for \n)
		# t=refseq.read(n)
		# t=t.strip('\n')
		# print t
		refseq.seek(jj1)
		t=''
		jj=0
		for jj in range(0,n):
			tt=refseq.read(1)
			if tt != '\n':
				t+=tt
		
		#######################################################################################
		# Get reference base at poi:
		#######################################################################################
	
		#tf=open ('t.txt', 'wb')
		#tf.write(t)
		#tf.close()
		
		kpoi=poi-j1
		#print ' poi=%s, kpoi =%s ' % (poi, kpoi)
		refbasepoi = t[kpoi:kpoi+1]
		#print ' refbasepoi = %s' % (refbasepoi)
		# sys.exit()
	
		#######################################################################################
		# Get context at poi:
		#######################################################################################

		# 2011_02_04
		### change the following line in pibase_bamref too! ###
		##lt=len(t)-1
		lt=len(t)
		
		#print min(kpoi+nflank+1,lt), kpoi+nflank+1, lt
		# print 'lt: %s' % (lt)
		ctxt1= t[max(0,kpoi-nflank):max(0,kpoi)]
		
		### change the following line in pibase_bamref too! ###
		##ctxt2 = t[min(kpoi+1,lt):min(kpoi+nflank,lt)]
		ctxt2 = t[min(kpoi+1,lt):min(kpoi+nflank+1,lt)]
		
		#print len(ctxt2), kpoi+1
		ctxt = ctxt1+'['+refbasepoi+']' + ctxt2 + '\n'
		
		#######################################################################################
		# write to fasta output file:
		#######################################################################################

		c='>%s:%s_%s\n' % (chrom, poi+1, refseqfile)
		out.write (c)
		
		out.write (ctxt1[0:100]+'\n')
		out.write (ctxt1[100:200]+'\n')
		out.write (ctxt1[200:300]+'\n')
		out.write (ctxt1[300:400]+'\n')
		out.write (ctxt1[400:500]+'\n')

		out.write ('['+refbasepoi+']\n')
		
		out.write (ctxt2[0:100]+'\n')
		out.write (ctxt2[100:200]+'\n')
		out.write (ctxt2[200:300]+'\n')
		out.write (ctxt2[300:400]+'\n')
		out.write (ctxt2[400:500]+'\n')
		
		#######################################################################################
		# Get next position of interest from table file
		#######################################################################################

		line=tabf.readline()
		if len(line) == 0:
			# EOF reached
			break
		line=line.strip()
		
		# Get chromosome_name and poi (0-based) from tab-separated file:
		try:
			cols = line.split('\t')
			chrom = cols[0]
			poi = int(cols[1].strip()) - 1
		except:
			print '*** Error: Bad format in line of file:  %s\n***        Line is:\n%s\n' % (tablefile,line)
			sys.exit('***        Line should be: <chromosome_name> TAB <1_based_position_in_chrom>')

	#######################################################################################################
	# End of WHILE loop over all positions of interest
	#######################################################################################################

	out.close()

###########################################################################################################
#
# the actual main
#
###########################################################################################################

###########################################################################################################
# set resource limits:
###########################################################################################################

setlimits()

###########################################################################################################
# Get command line arguments
#
# pibase_ref  infile  ref outfile  [nflank]
###########################################################################################################

nargs = len(sys.argv)
if nargs != 4:
	print pib_ref.__doc__
	sys.exit()

# defaults
nflank=500

tablefile = sys.argv[1]
refseqfile = sys.argv[2]
outfile = sys.argv[3]

###########################################################################################################
# call main procedure
###########################################################################################################

pib_ref (refseqfile, tablefile, outfile, nflank)
print('*** pibase_ref completed successfully')


