#!/usr/bin/env python
##!/usr/bin/python

import sys
import time
import resource

__author__="mforster"
__date__ ="$Mar 28, 2011 1:00 PM$"

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_tosnpactsown (out, chrom, chr, pos1, ref, bestgen):
	###########################################################################
	# generate a line in snpacts 'own' format and write into filehandle out
	# col 1 - chromosome name ( -pchr 1 )
	# col 2 - 1-based position ( -sp 1 -psnp 2 )
	# col 3 - ref allele ( -prb 3 )
	# col 4 - best allele ( -pbb 4 )
	# col 5 - second best allele ( -psb 5 )
	# col 6 - chromosome name (pchr = 1)
	# snpacts-import needs to be invoked (from cmd line) as follows: 
	# snpact -ft own -hg hg18 -all (...above 6 switches...) -f file -t table
	###########################################################################
	line = chrom[chr] + '\t' + str(pos1) + '\t'  + ref + '\t'
	bestallele = ref
	secondallele = ref
	muttype = 'hom'
	if bestgen != '':
		bestallele = bestgen [:1]
		secondallele = bestgen [1:]
		if bestallele != secondallele:
			muttype = 'het'
	line += bestallele + '\t' + secondallele + '\t' + muttype + '\n'
	out.write(line)

def pib_tosnpacts1 (infile, inf, ic, maxcol, n, k):
	###########################################################################
	# called by pib_tosnpacts: to read and process a single line from infile
	###########################################################################
	line = inf.readline()
	# print line
	if len(line) == 0:
		# EOF
		if k == 1:
			return 0,0,'',''
		else:
			# pibase_diff
			return 0,0,'','',''
	cols=line.split('\t')
	ncols = len(cols)
	if ncols < maxcol:
		# end of table
		if k == 1:
			return 0,0,'',''
		else:
			# pibase_diff
			return 0,0,'','',''
	try:
		chr = int (cols[ ic[0] ])
		pos1 = int (cols[ ic[1]  ])
	except:
		print ' *** Error: Unable to read integer values (chrom column and pos column)'
		print ' Line:', line[:30]
		print ' cols[ic[0]], cols[pos_col]: ', cols[ic[0]], cols[ic[1]]
		sys.exit (' Details: chrom-column %s, pos-column %s, file %s, line %s' % (ic[0]+1, ic[1]+1, infile, n))
	ref = cols[ ic[2] ]
	if len(ref) != 1 or 'ACGTNacgtn'.find(ref)<0 :
		print ' *** Error: Bad reference base (not length 1) or not in ACGTNacgtn:'
		print ' Line:', line[:30]
		print ' ref: ', ref
		sys.exit (' file %s, line number %s' % (infile, n))
		
	bestgen = cols[ ic[3] ]		
	if len(bestgen) != 2: 
		if bestgen != '':
			print ' *** Warning: Bad BestGen (not length 2): BestGen = %s' % ( bestgen )
			# print ' Line:', line[:30]
			print ' *** File: %s   Line number: %s' % (infile, n)
			print ' *** Replacing with Reference (only in order to get snpacts locus annotation) and continuing normally.' 
		bestgen = ref+ref
	if k == 1:
		# print 'chr, pos1, ref, bestgen :', chr, pos1, ref, bestgen
		return chr, pos1, ref, bestgen
	else:
		# pibase_diff
		# print 'chr, pos1, ref, bestgen, #D:', chr, pos1, ref, bestgen, cols[0]
		return chr, pos1, ref, bestgen, cols[0]

def pib_tosnpacts (tfile, infile, outfile1, outfile2):
	'''        #######################################################################################
	#  pibase_tosnpacts.py  v1.4.5
	#######################################################################################
	#  Export a file from pibase_consensus, pibase_fisherdiff, or pibase_diff, 
	#  which can be imported into snpacts and annotated. 
	#  After annotation and export (using outputact): use pibase_fromsnpacts to merge 
	#  the annotations into the pibase_consensus, pibase_fisherdiff, or pibase_diff file.
	#######################################################################################
	#  pibase_tospnacts [-n tfile] <in> <out1> [out2]
	# 
	#  []=optional
	#
	#  [-t tfile] file with contig numbering/name table in header
	#             (required, if the <in> file is a pibase_fisherdiff file,
	#             because the pibase_fisherdiff file has a shortened file header
	#             without the conversion table)
	#
	#  in         file to be converted (from pibase_consensus, pibase_fisherdiff, ...)
	# 
	#  out1       output file (for input into snpacts, in 'own' format)
	#
	#  [out2]     2nd output file, i.e. if pibase_fisherdiff or pibase_diff
	#######################################################################################
	#  Author: Michael Forster, m.forster@ikmb.uni-kiel.de         
	#  Date: 8 August 2012
	#######################################################################################
	'''
	
	print ' Tfile:', tfile 
	print ' Infile:', infile 
	print ' Outfile1:', outfile1 
	print ' Outfile2:', outfile2 
	
	
	############################################################################
	# open tfile and find contig numbering/naming table in header 
	############################################################################
	
	inf = open (tfile, 'r')
	while True:
		line = inf.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: There is no table header line containing "# Mapping of pibase-chromosome-number to chromosome/contig name:"'
			sys.exit (' in file %s' % (tfile))
		if line.find('# Mapping of pibase-chromosome-number to chromosome/contig name:')> -1 :
			line = inf.readline()
			break
	
	############################################################################
	# read contig numbering/naming table into dictionary 
	############################################################################
	
	chrom = {}
	n = 0
	while True:
		line = inf.readline().strip()
		cols = line.split('\t')
		if len(cols) != 3 :
			break
		n += 1
		if int(cols[1]) != n :
			print ' *** Error: There is no continuous contig numbering'
			sys.exit (' in file %s' % (infile))
		chrom[n] = cols[2]
	if n == 0:
		print ' *** Error: There is no contig numbering/name table in the header lines'
		sys.exit (' in file %s' % (infile))
	
	inf.close()
	
	############################################################################
	# open infile and test for file type 
	############################################################################
	
	ft = -1
	
	inf = open (infile, 'r')
	while True:
		line = inf.readline()
		if len(line) == 0:
			# EOF
			print ' *** Error: There is no table header line containing "RefSeqContext<TAB>GC[%]"'
			sys.exit (' in file %s' % (infile))
		if line.find('RefSeqContext\tGC[%]')> -1 :
			header = line
			break
	if header.startswith('# C\tPos[1]'):
		# pibase_consensus
		ft = 1
	elif header.startswith('#D\tD_error\tp.med'):
		# pibase_fisherdiff or pibase_fisherdiff_bestchoice
		if header.find('\tChr') > -1 :
			# pibase_fisherdiff_bestchoice
			ft = 2
		else:
			ft = 3
	elif header.startswith('#D\tC'):
		# pibase_diff
		ft = 3
	else:
		print ' *** Error: The table header line is inconsistent with a valid input file format.'
		print ' *** File: %s' % (infile)
		print ' *** Table header: ', header [:30]
		sys.exit (' *** Allowed file formats: pibase_consensus, pibase_fisherdiff, or pibase_diff output.')
	
	inf.close()

	if (ft ==2 or ft == 3) and outfile2 == '':
		print pib_tosnpacts.__doc__
		sys.exit()
		
	############################################################################
	# find columns chr, pos, ref, bestgen
	############################################################################
	
	header = header.strip()
	# print header
	cols=header.split('\t')
	c=[]
	if ft == 1:
		# pibase_consensus
		c.append('# C')
	elif ft == 2:
		# pibase_fisherdiff_bestchoice
		c.append('Chr')
	else:
		# pibase_fisherdiff or pibase_diff
		c.append('C')
	c.append('Pos[1]')
	c.append('Ref')
	c.append('BestGen')
	ic=[]
	maxcol=0
	for i in c:
		# print i
		n=0
		for col in cols:
			if i==col:
				# print 'found '+col+' in column '+str(n)
				ic.append(n)
				maxcol=max(maxcol,n)
				break
			n+=1
	if len(ic) != len(c):
			print ' *** Error: 1 or more columns of interest not found in header line'
			print  c
			print ic
			print ' *** header line = ' + header
			sys.exit (' *** Aborted' )
	# print 'ic',ic


	############################################################################
	# open infile and outfile(s) and write brief header lines to outfile(s):
	############################################################################
	
	inf = open (infile, 'r')
	out1 = open (outfile1, 'w')
	
	outtime = time.strftime("%Y_%m_%d__%H:%M:%S") + "\n"
	out1.write ('# File for snpact import created by pibase_tosnpacts v1.4.5 on '+ outtime)
	out1.write ('# from: %s\n' %(infile))
	out1.write ('#Chr\tPos[1]\tRef\tBest\t2ndBest\n')

	if ft >= 2:
		out2 = open (outfile2, 'w')
		out2.write ('# File 2 for snpact import created by pibase_tosnpacts v1.4.5 on '+ outtime)
		out2.write ('# from: %s\n' %(infile))
		out2.write ('#Chr\tPos[1]\tRef\tBest\t2ndBest\n')

	############################################################################
	# read infile comment lines up to and including table header 
	############################################################################

	n = 0
	while True:
		n += 1
		line = inf.readline()
		if line.find('RefSeqContext\tGC[%]')> -1 :
			break
	# print line

	############################################################################
	# loop over remaining lines of infile
	############################################################################
	
	if ft == 1:
		# pibase_consensus
		while True:
			n += 1
			# read and process a single line from infile
			chr, pos1, ref, bestgen = pib_tosnpacts1 (infile, inf, ic, maxcol, n, 1)
			if (chr, pos1, ref, bestgen) ==  (0,0,'','') :
				break
			if bestgen != '':
				# generate a line in snpacts 'own' format and write into outfile
				pib_tosnpactsown (out1, chrom, chr, pos1, ref, bestgen)
	elif ft == 2:
		# pibase_fisherdiff
		while True:
			n += 1
			# read and process a single line from infile
			chr, pos1, ref, bestgen = pib_tosnpacts1 (infile, inf, ic, maxcol, n, 1)
			if (chr, pos1, ref, bestgen) ==  (0,0,'','') :
				break
			if bestgen != '':
				# generate a line in snpacts 'own' format and write into outfile 1
				pib_tosnpactsown (out1, chrom, chr, pos1, ref, bestgen)
			
			n += 1
			# read and process a single line from infile
			chr, pos1, ref, bestgen = pib_tosnpacts1 (infile, inf, ic, maxcol, n, 1)
			if (chr, pos1, ref, bestgen) ==  (0,0,'','') :
				break
			if bestgen != '':
				# generate a line in snpacts 'own' format and write into outfile 2
				pib_tosnpactsown (out2, chrom, chr, pos1, ref, bestgen)
	elif ft == 3:
		# pibase_diff
		while True:
			n += 1
			# read and process a single line from infile
			chr, pos1, ref, bestgen, d = pib_tosnpacts1 (infile, inf, ic, maxcol, n, 0)
			if (chr, pos1, ref, bestgen, d) ==  (0,0,'','','') :
				break
			if bestgen != '':
				# generate a line in snpacts 'own' format and write into outfile 1
				pib_tosnpactsown (out1, chrom, chr, pos1, ref, bestgen)
			
			if d != '=' :
				n += 1
				# read and process a single line from infile
				chr, pos1, ref, bestgen, d = pib_tosnpacts1 (infile, inf, ic, maxcol, n, 0)
				if (chr, pos1, ref, bestgen, d) ==  (0,0,'','','') :
					break
			if bestgen != '':
				# generate a line in snpacts 'own' format and write into outfile 2
				pib_tosnpactsown (out2, chrom, chr, pos1, ref, bestgen)
	out1.close()
	if ft == 2 or ft == 3:
		out2.close()
	inf.close()

##############################################
# the actual main
##############################################

if len(sys.argv) < 3:
	print pib_tosnpacts.__doc__
	sys.exit()
else:
	if sys.argv[1] == '-t':
		# command line includes an additional translation table file 
		# (for converting the pibase-contig-numbers to the contig-names)
		if len(sys.argv) < 5:
			print pib_tosnpacts.__doc__
			sys.exit()
		tfile = sys.argv[2]
		infile = sys.argv[3]
		outfile1 = sys.argv[4]
		if len(sys.argv) == 6:
			outfile2 = sys.argv[5]
		else:
			outfile2 = ''
	else:
		# no additional translation table file given
		# (the table is already incorporated in the header of
		# pibase_bamref, pibase_consensus, or pibase_flagsnp files)
		
		infile = sys.argv[1]
		tfile = infile
		outfile1 = sys.argv[2]
		if len(sys.argv) == 4:
			print pib_tosnpacts.__doc__
			print ' *** Error: inconsistent number of command line arguments: you must also specify a tfile if you specified 2 output files'
			print ' *** in = %s' % (infile)
			print ' *** out1 = %s' % (outfile1)
			print ' *** out2 = %s' % (sys.argv[3])
			sys.exit()
		else:
			outfile2 = ''
	
setlimits()

pib_tosnpacts (tfile, infile, outfile1, outfile2)
print('Done')
